#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <float.h>
#include "bds.h"
#include "gdstool.h"
#include "gds.h"
#include "gribw.h"

/*
 * c2pgrb -- A GRIB Cartesian to polar converter.
 *
 * Compile: make -f p2cgrb.make c2pgrb
 * 
 * Usage: type c2pgrb without arguments
 * 
 * Requires: libgribw
 *
 * oyvind.breivik@met.no
 *
 */

/* Constants */
#define VERSION "c2pgrb v1.0, 2004m08d05, oyvind.breivik@met.no"
#define USAGE "Usage: %s u.grb v.grb polar.grb [[-]magnitude-par dir-par]\n"
#define N 100000
#define WDIR 31
#define WIND 32 /* International table wind direction and speed [default] */

int main(int argc, char **argv) {

   /* Arrays */
   unsigned char *pds, *gds, *bms, *bds;
   float *ffarr, *ddarr, *uarr, *varr; /* tmp work arrays */

   /* Vars */
   long int len_grib, upos=0L, vpos=0L; /* GRIB file position counters */

   int nrec;      /* record counter */
   int nxny;      /* grid dimension */
   int k;         /* index */
   int magnitude=WIND, direction=WDIR; /* parameter codes for magnitude & dir
                                          default is wind speed and wind dir */
   int sgn=-1;     /* sign: -1 means direction is "coming from", the
                      meteorological convention [default] */

   float pi, rad; /* yes, good old PI */

   FILE *uin, *vin, *output;

   /* Initialize */
   pi = 4*atan(1.0);
   rad = pi/180.0;

   /* Help line */
   if (argc == 1) {
      fprintf(stderr, "%s\n", VERSION);
      fprintf(stderr,"Transforms GRIB records from (east,north)-components to (magnitude,direction).\n");
      fprintf(stderr,
"Note: the angle is measured in degrees clockwise from north [0,360), \nin the direction from which the vector is pointing, i.e., the coming-from or \nmeteorological convention. To revert this, add a negative sign to either of \nthe two parameters.\n");
      fprintf(stderr,"Example:\n   %s u.grb v.grb polar.grb -48 47\n",argv[0]);
      fprintf(stderr,
"The new file polar.grb contains the speed and direction of the current vector\n with parameter nos 48 and 47 [optional, 32 and 31 are default].\nDirection is measured in the going-to or oceanographic convention.\n");
      exit(8);
   }

   /* Open files */
   if (argc < 4) {
      fprintf(stderr, USAGE, argv[0]);
      exit(8);
   }
   if ((uin = fopen(argv[1],"rb")) == NULL) {
      fprintf(stderr,"Could not open file: %s\n", argv[1]); 
      exit(7);                                              
   }                                                         
   if ((vin = fopen(argv[2],"rb")) == NULL) {
      fprintf(stderr,"Could not open file: %s\n", argv[2]); 
      exit(7);                                              
   }                                                         
   if ((output = fopen(argv[3],"wb")) == NULL) {             
      fprintf(stderr,"Could not open file: %s\n", argv[3]); 
      exit(7);                                              
   }

   if (argc > 4) {
      magnitude = atoi(argv[4]);
      direction = atoi(argv[5]);
      if (magnitude < 0 || direction < 0) {
         sgn=1;
         magnitude=abs(magnitude);
         direction=abs(direction);
      }
   }
   
   /* Preallocate BIG arrays for speed */
   ffarr = (float *) malloc(N * sizeof(float));
   ddarr = (float *) malloc(N * sizeof(float));
   uarr  = (float *) malloc(N * sizeof(float));
   varr  = (float *) malloc(N * sizeof(float));
   

 /*** Loop over input records ***/

   for (nrec=0; 1; nrec+=2) {


    /*** Read east-component ***/

      len_grib = rd_grib_msg(uin, upos, &pds, &gds, &bms, &bds);
      if (len_grib <= 0) break;
      upos += len_grib;

      /* Unpack east-component */
      unpk_bds(uarr, pds, gds, bms, bds, nxny);

      /* Reallocate arrays if necessary */
      nxny = get_nxny(pds, gds, bms, bds);
      if (nxny > N) {
         free(ffarr);
         free(ddarr);
         free(uarr);
         free(varr);
         ffarr =  (float *) malloc((nxny) * sizeof(float));
         if (ffarr == NULL) {
            fprintf(stderr,"Memory allocation failed\n");
         }
         ddarr =  (float *) malloc((nxny) * sizeof(float));
         if (ddarr == NULL) {
            fprintf(stderr,"Memory allocation failed\n");
         }
         uarr =  (float *) malloc((nxny) * sizeof(float));
         if (uarr == NULL) {
            fprintf(stderr,"Memory allocation failed\n");
         }
         varr =  (float *) malloc((nxny) * sizeof(float));
         if (varr == NULL) {
            fprintf(stderr,"Memory allocation failed\n");
         }
      } /* end if nxny */


    /*** Read north-component ***/

      len_grib = rd_grib_msg(vin, vpos, &pds, &gds, &bms, &bds);
      if (len_grib <= 0) break;
      vpos += len_grib;

      /* Unpack north-component */
      unpk_bds(varr, pds, gds, bms, bds, nxny);


    /*** Convert to magnitude and direction (clockwise rel north) ***/

      for (k = 0; k<nxny; k++) {
         ffarr[k] = sqrt(uarr[k]*uarr[k]+varr[k]*varr[k]);
         ddarr[k] = atan2(sgn*uarr[k],sgn*varr[k])/rad;
         /* Angles [0,360) */
         ddarr[k] = ddarr[k] < 0 ? ddarr[k] + 360.0 : ddarr[k] ;
      } /* end for */


    /*** Write magnitude GRIB record ***/

      /* Remake product description section, PDS */
      pds[8] = magnitude; /* PDS octet 9, magnitude [wind speed] */

      /* ECMWF style? */
      if (BDS_BinScale(bds)) {
         set_BDSMinBits(BDS_NumBits(bds));
         set_BDSMaxBits(BDS_NumBits(bds));
      }

      /* Remake binary data section, BDS */
      bds = mk_BDS(pds, ffarr, nxny); /* magnitude */

      /* Write GRIB record */
      wrt_grib_msg(output, pds, gds, bms, bds); /* magnitude */


    /*** Write direction GRIB record (clockwise rel north) ***/

      /* Remake product description section, PDS */
      pds[8] = direction; /* PDS octet 9, direction [wind dir] */

      /* ECMWF style? */
      if (BDS_BinScale(bds)) {
         set_BDSMinBits(BDS_NumBits(bds));
         set_BDSMaxBits(BDS_NumBits(bds));
      }

      /* Remake binary data section, BDS */
      bds = mk_BDS(pds, ddarr, nxny); /* direction */

      /* Write GRIB record */
      wrt_grib_msg(output, pds, gds, bms, bds); /* direction */

   } /* end for nrec */

   /* Close files */
   fclose(uin);
   fclose(vin);
   fclose(output);

   /* Clean up */
   if (ffarr != NULL) {
     free(ffarr);
     ffarr = NULL;
   }
   if (ddarr != NULL) {
     free(ddarr);
     ddarr = NULL;
   }
   if (uarr != NULL) {
     free(uarr);
     uarr = NULL;
   }
   if (varr != NULL) {
     free(varr);
     varr = NULL;
   }

   /* Output */
   if (nrec > 1) {
      printf("Wrote %d records\n", nrec);
   }
   else if (nrec == 1) {
      printf("Wrote one record\n");
   }
   else {
      fprintf(stderr,"ERROR: no records written\n");
      exit(4);
   } /* end if */

   return 0;
} /* end main */
