#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <float.h>
#include "bds.h"
#include "gdstool.h"
#include "gds.h"
#include "gribw.h"

/*
 * p2cgrb -- A GRIB polar to Cartesian converter.
 *
 * Compile: make -f p2cgrb.make
 * 
 * Usage: type p2cgrb without arguments
 * 
 * Requires: libgribw
 *
 * oyvind.breivik@met.no
 *
 */

/* Constants */
#define VERSION "p2cgrb v1.0, 2004m08d05, oyvind.breivik@met.no"
#define USAGE "Usage: %s speed.grb dir.grb uv.grb [[-]u-par v-par]\n"
#define N 100000
#define UGRD 33
#define VGRD 34 /* International table u and v wind components (defaults) */

int main(int argc, char **argv) {

   /* Arrays */
   unsigned char *pds, *gds, *bms, *bds;
   float *ffarr, *ddarr, *uarr, *varr; /* tmp work arrays */

   /* Vars */
   long int len_grib, ffpos=0L, ddpos=0L; /* GRIB file position counters */

   int nrec;      /* record counter */
   int nxny;      /* grid dimension */
   int k;         /* index */
   int ugrd=UGRD, vgrd=VGRD; /* parameter codes for u and v components */
   int sgn=-1;     /* sign: -1 means direction is "coming from", the
                      meteorological convention [default] */

   float pi, rad; /* yes, good old PI */

   FILE *ffin, *ddin, *output;

   /* Initialize */
   pi = 4*atan(1.0);
   rad = pi/180.0;

   /* Help line */
   if (argc == 1) {
      fprintf(stderr, "%s\n", VERSION);
      fprintf(stderr,"Transforms GRIB records from (magnitude,direction) to (east,north)-components.\n");
      fprintf(stderr,
"Note: the angle is assumed to be in degrees clockwise from north, \nin the direction from which the vector is pointing, i.e., the coming-from or \nmeteorological convention. To revert this, add a negative sign to either of \nthe two parameters.\n");
      fprintf(stderr,"Example:\n   %s speed.grb dir.grb uv.grb -49 50\n",argv[0]);
      fprintf(stderr,
"The new file uv.grb contains the east and north components of the current\nvector with parameter nos 49 and 50 [optional, 33 and 34 are default].\nDirection is measured in the going-to or oceanographic convention.\n");
      exit(8);
   }

   /* Open files */
   if (argc < 4) {
      fprintf(stderr, USAGE, argv[0]);
      exit(8);
   }
   if ((ffin = fopen(argv[1],"rb")) == NULL) {
      fprintf(stderr,"Could not open file: %s\n", argv[1]); 
      exit(7);                                              
   }                                                         
   if ((ddin = fopen(argv[2],"rb")) == NULL) {
      fprintf(stderr,"Could not open file: %s\n", argv[2]); 
      exit(7);                                              
   }                                                         
   if ((output = fopen(argv[3],"wb")) == NULL) {             
      fprintf(stderr,"Could not open file: %s\n", argv[3]); 
      exit(7);                                              
   }

   if (argc > 4) {
      ugrd = atoi(argv[4]);
      vgrd = atoi(argv[5]);
      if (ugrd < 0 || vgrd < 0) {
         sgn=1;
         ugrd=abs(ugrd);
         vgrd=abs(vgrd);
      }
   }
   
   /* Preallocate BIG arrays for speed */
   ffarr = (float *) malloc(N * sizeof(float));
   ddarr = (float *) malloc(N * sizeof(float));
   uarr  = (float *) malloc(N * sizeof(float));
   varr  = (float *) malloc(N * sizeof(float));
   

 /*** Loop over input records ***/

   for (nrec=0; 1; nrec+=2) {


    /*** Read direction ***/

      len_grib = rd_grib_msg(ddin, ddpos, &pds, &gds, &bms, &bds);
      if (len_grib <= 0) break;
      ddpos += len_grib;

      /* Unpack direction */
      unpk_bds(ddarr, pds, gds, bms, bds, nxny);

      /* Reallocate arrays if necessary */
      nxny = get_nxny(pds, gds, bms, bds);
      if (nxny > N) {
         free(ffarr);
         free(ddarr);
         free(uarr);
         free(varr);
         ffarr =  (float *) malloc((nxny) * sizeof(float));
         if (ffarr == NULL) {
            fprintf(stderr,"Memory allocation failed\n");
         }
         ddarr =  (float *) malloc((nxny) * sizeof(float));
         if (ddarr == NULL) {
            fprintf(stderr,"Memory allocation failed\n");
         }
         uarr =  (float *) malloc((nxny) * sizeof(float));
         if (uarr == NULL) {
            fprintf(stderr,"Memory allocation failed\n");
         }
         varr =  (float *) malloc((nxny) * sizeof(float));
         if (varr == NULL) {
            fprintf(stderr,"Memory allocation failed\n");
         }
      } /* end if nxny */


    /*** Read speed ***/

      len_grib = rd_grib_msg(ffin, ffpos, &pds, &gds, &bms, &bds);
      if (len_grib <= 0) break;
      ffpos += len_grib;

      /* Unpack speed */
      unpk_bds(ffarr, pds, gds, bms, bds, nxny);


    /*** Convert to east and north components ***/

      for (k = 0; k<nxny; k++) {
         uarr[k] = sgn*ffarr[k]*sin(ddarr[k]*rad);
         varr[k] = sgn*ffarr[k]*cos(ddarr[k]*rad);
      } /* end for */


    /*** Write east component GRIB record ***/

      /* Remake product description section, PDS */
      pds[8] = ugrd; /* PDS octet 9, east component */

      /* ECMWF style? */
      if (BDS_BinScale(bds)) {
         set_BDSMinBits(BDS_NumBits(bds));
         set_BDSMaxBits(BDS_NumBits(bds));
      }

      /* Remake binary data section, BDS */
      bds = mk_BDS(pds, uarr, nxny); /* east component */

      /* Write GRIB record */
      wrt_grib_msg(output, pds, gds, bms, bds); /* ugrd, east comp.  */


    /*** Write north component GRIB record ***/

      /* Remake product description section, PDS */
      pds[8] = vgrd; /* PDS octet 9, north component */

      /* ECMWF style? */
      if (BDS_BinScale(bds)) {
         set_BDSMinBits(BDS_NumBits(bds));
         set_BDSMaxBits(BDS_NumBits(bds));
      }

      /* Remake binary data section, BDS */
      bds = mk_BDS(pds, varr, nxny); /* north component */

      /* Write GRIB record */
      wrt_grib_msg(output, pds, gds, bms, bds); /* vgrd, north comp. */

   } /* end for nrec */

   /* Close files */
   fclose(ffin);
   fclose(ddin);
   fclose(output);

   /* Clean up */
   if (ffarr != NULL) {
     free(ffarr);
     ffarr = NULL;
   }
   if (ddarr != NULL) {
     free(ddarr);
     ddarr = NULL;
   }
   if (uarr != NULL) {
     free(uarr);
     uarr = NULL;
   }
   if (varr != NULL) {
     free(varr);
     varr = NULL;
   }

   /* Output */
   if (nrec > 1) {
      printf("Wrote %d records\n", nrec);
   }
   else if (nrec == 1) {
      printf("Wrote one record\n");
   }
   else {
      fprintf(stderr,"ERROR: no records written\n");
      exit(4);
   } /* end if */

   return 0;
} /* end main */
